@@ -2,7 +2,7 @@ package Callback::Frame;
 
 use strict;
 
-our $VERSION = '0.1';
+our $VERSION = '0.2';
 
 require Exporter;
 use base 'Exporter';
@@ -87,10 +87,44 @@ sub frame {
 
     local $top_of_stack = $new_frame;
 
-    $code = generate_local_wrapped_code($top_of_stack, $code);
+    my $frame_i = $top_of_stack;
 
     my $val = eval {
+      ## Find applicable local vars
+
+      my $local_refs = {};
+      my $temp_copies = {};
+
+      for(; $frame_i; $frame_i = $frame_i->{down}) {
+        next unless exists $frame_i->{locals};
+        foreach my $k (keys %{$frame_i->{locals}}) {
+          next if exists $local_refs->{$k};
+          $local_refs->{$k} = \$frame_i->{locals}->{$k};
+        }
+      }
+
+      ## Backup local vars
+
+      foreach my $var (keys %$local_refs) {
+        no strict qw/refs/;
+        $temp_copies->{$var} = $$var;
+        $$var = ${$local_refs->{$var}};
+      }
+
+      ## Install code that will restore local vars
+
+      scope_guard {
+        foreach my $var (keys %$local_refs) {
+          no strict qw/refs/;
+          ${$local_refs->{$var}} = $$var;
+          $$var = $temp_copies->{$var};
+        }
+      };
+
+      ## Actually run the callback
+
       $@ = $orig_error;
+
       $code->(@_);
     };
 
@@ -160,41 +194,6 @@ sub generate_trace {
 }
 
 
-sub generate_local_wrapped_code {
-  my ($frame_i, $code) = @_;
-
-  my $local_refs = {};
-
-  for(; $frame_i; $frame_i = $frame_i->{down}) {
-    next unless exists $frame_i->{locals};
-    foreach my $k (keys %{$frame_i->{locals}}) {
-      next if exists $local_refs->{$k};
-      $local_refs->{$k} = \$frame_i->{locals}->{$k};
-    }
-  }
-
-  return $code unless keys %{$local_refs};
-
-  my $codestr = 'sub { no strict qw/refs/;';
-
-  foreach my $k (keys %$local_refs) {
-    $codestr .= qq{ local \$$k = \${\$local_refs->{q{$k}}}; };
-  }
-
-  $codestr .= ' scope_guard { ';
-  foreach my $k (keys %$local_refs) {
-    $codestr .= qq{ \${\$local_refs->{q{$k}}} = \$$k; };
-  }
-  $codestr .= ' }; ';
-
-  $codestr .= ' return $code->(@_); } ';
-
-  my $new_code = eval($codestr);
-  die "local frame wrapper compile failure: $@ ($codestr)\n" if $@;
-
-  return $new_code;
-}
-
 
 1;
 
@@ -254,6 +253,8 @@ will print C<123> even though C<$var> is no longer in scope when the callback is
 
 Sometimes people call these anonymous functions that reference variables in their surrounding lexical scope "closures". Whatever you call them, they are essential for convenient and efficient asynchronous programming. 
 
+The Callback::Frame philosophy is that for many applications we really like callback style and don't wish to run (or pretend to run) multiple blocking tasks at once. Callback::Frame just tries to simplify the management of dynamic environments (defined below) while leaving callback style alone.
+
 
 =head1 DESCRIPTION
 
@@ -308,7 +309,7 @@ Now we see the desired error message:
 
     Oops: some error at fixed.pl line 8.
 
-We created two framed to accomplish this: A root frame to contain the exception handler, and another frame to use as a callback. Because the callback frame is created while the root frame is running, the callback will preserve the dynamic environment of the root frame.
+We created two frames to accomplish this: A root frame to contain the exception handler, and a nested frame to use as a callback. Because the nested callback frame is created while the root frame is executing, the callback will preserve the dynamic environment (including the exception handler) of the root frame.
 
 
 
@@ -318,7 +319,7 @@ This module exports two functions, C<frame> and C<fub>.
 
 C<frame> is the general interface and requires at least a C<code> argument which should be a coderef (a function or a closure). It will return another coderef that "wraps" the coderef you passed in. When this codref is run, it will re-instate the dynamic environment that was present when the frame was created, and then run the coderef that you passed in as C<code>.
 
-B<IMPORTANT NOTE>: All callbacks that may be invoked outside the dynamic environment of the current frame should be wrapped in a new C<frame> or C<fub> call so that the dynamic environment will be correctly re-applied when the callback is invoked.
+B<IMPORTANT NOTE>: All callbacks that may be invoked outside the dynamic environment of the current frame should be created with C<frame> or C<fub> so that the dynamic environment will be correctly re-applied when the callback is invoked.
 
 C<fub> is a wrapper around frame and exists to simplify converting existing callbacks into Callback::Frame enabled code. For example, given the following L<AnyEvent> statement:
 
@@ -340,6 +341,8 @@ You should never need to, but the internal frame stack can be accessed at C<$Cal
 
 =head1 NESTING AND STACK-TRACES
 
+Callback::Frame tries to make adding error handling support to an existing asynchronous application as easy as possible by not forcing you to pass extra parameters around. It should also make lifer easier because as a side effect of adding error checking it also can be made to produce detailed and useful "stack traces" that track the callback history of some connection or transaction.
+
 Frames can be nested. When an exception is raised, the most deeply nested C<catch> handler is invoked. If this handler itself throws an error, the next most deeply nested handler is invoked with the new exception but the original stack trace. If the last C<catch> handler re-throws the error, the error will simply be thrown to your application from wherever the callback happened to be called (probably not what you want).
 
 When a C<catch> handler is called, not only is C<$@> set, but also a stack-trace string is passed in as the first argument. All frames will be listed in this stack-trace, starting with the most deeply nested frame.
@@ -358,17 +361,18 @@ In the same way that using C<catch> as described above preserves the dynamic env
 
 Technically, C<local> maintains the dynamic environment of B<bindings>. Although perl coders often ignore the distinction, lisp coders are careful to distinguish between variables and bindings. See, when a lexical binding is created, it is there "forever" -- or at least until it is no longer reachable by your program according to the rules of lexical scoping. Therefore, lexical variables are statically mapped to bindings and it is redundant to distinguish between lexical variables and bindings.
 
-However, with dynamic variables the same variable can refer to different bindings at different times. That's why they are called "dynamic" variables and lexical variables are called "static" variables.
+However, with dynamic variables the same variable can refer to different bindings at different times. That's why they are called "dynamic" and lexical variables are called "static".
 
 Because any code in any file, function, or package can access a dynamic variable, they are the opposite of local. They are global. However, the bindings are only global for a little while at a time. After a while they will go out of scope and then they are no longer visible at all. Or sometimes they will get "shadowed" by some other binding and will come back again later.
 
-OK, to make this concrete, after running this bit of code the binding containing C<2> is lost forever:
+OK, to make this concrete, here the binding containing C<2> is lost forever:
 
     our $foo = 1;
     my $cb;
 
     {
-      local $foo = 2;
+      local $foo;
+      $foo = 2;
       $cb = sub {
         return $foo;
       };
@@ -386,16 +390,16 @@ Here's a way to "fix" that using Callback::Frame:
     frame(local => __PACKAGE__ . '::foo',
           code => sub {
       $foo = 2;
-      $cb = frame(code => sub {
+      $cb = fub {
         return $foo;
-      });
+      };
     })->();
 
     say $foo;     # 1
     say $cb->();  # 2  <- hooray!
     say $foo;     # 1
 
-Don't be fooled into thinking that this is a lexical binding though. While the variable C<$foo> points to the binding containing C<2>, all parts of the program will see this binding as demonstrated in the following example:
+Don't be fooled into thinking that this is a lexical binding though. While the callback C<$cb> is executing, all parts of the program will see the C<2> binding:
 
     our $foo = 1;
     my $cb;
@@ -407,14 +411,14 @@ Don't be fooled into thinking that this is a lexical binding though. While the v
     frame(local => __PACKAGE__ . '::foo',
           code => sub {
       $foo = 2;
-      $cb = frame(code => sub {
+      $cb = fub {
         return global_foo_getter();
-      });
+      };
     })->();
 
     say $foo;     # 1
     say $cb->();  # 2  <- still 2
-    $foo;         # 1
+    say $foo;     # 1
 
 
 You can install multiple local variables in the same frame:
@@ -425,7 +429,7 @@ You can install multiple local variables in the same frame:
 
 Note that if you have both C<catch> and C<local> elements in a frame, in the event of an error the local bindings will B<not> be present inside the C<catch> handler (use a nested frame if you need this).
 
-All variable names must be fully package qualified. The easiest way to do this for variables in your current package is to use the ugly C<__PACKAGE__ .> technique.
+All variable names must be fully package qualified. The best way to do this for variables in your current package is to use the ugly C<__PACKAGE__> technique.
 
 Objects stored in local bindings managed by Callback::Frame will not be destroyed until all references to the frame-wrapped callback that contains the binding, along with all references to any deeper frames, are freed.
 
@@ -435,11 +439,9 @@ Objects stored in local bindings managed by Callback::Frame will not be destroye
 
 L<The Callback::Frame github repo|https://github.com/hoytech/Callback-Frame>
 
-The Callback::Frame philosophy is that we actually like callback style and just want to simplify the reseting of dynamic environments as necessary. This module should make adding error handling support to an existing asynchronous application as easy as possible (it shouldn't force you to pass extra parameters around). Finally, this module should make lifer easier because as a side effect of adding error checking it should also produce detailed and useful stack traces.
-
 This module's C<catch> syntax is of course modeled after "normal language" style exception handling as implemented by L<Try::Tiny> &c.
 
-This module depends on C<Guard> to maintain the C<$Callback::Frame::active_frames> datastructure, and to ensure that even when exceptions are thrown, C<local> binding updates aren't lost.
+This module depends on L<Guard> to maintain the C<$Callback::Frame::active_frames> datastructure and to ensure that C<local> binding updates aren't lost even when exceptions or other non-local returns occur.
 
 L<AnyEvent::Debug> is a very interesting and useful module that provides an interactive debugger for AnyEvent applications and uses some of the same techniques that Callback::Frame does. L<AnyEvent::Callback> and L<AnyEvent::CallbackStack> sort of solve the dynamic error handler problem. Unlike these modules, Callback::Frame is not related at all to L<AnyEvent>, except that it happens to be useful in AnyEvent libraries and applications (among other things).
 
@@ -457,9 +459,6 @@ L<UNWIND-PROTECT vs. Continuations|http://www.nhplace.com/kent/PFAQ/unwind-prote
 
 For now, C<local> bindings can only be created in the scalar namespace. Also, none of the other nifty things that L<local> can do (like localising a hash table value) are supported yet.
 
-The C<local> implementation is currently pretty inefficient. It evals a string that returns a sub whenever a frame is entered. Fortunately, perl's compiler is really fast. Also, this overhead is not there at all when you are only using the C<catch> functionality which I anticipate to be the most common usage pattern.
-
-
 
 
 =head1 AUTHOR
@@ -477,9 +476,6 @@ This module is licensed under the same terms as perl itself.
 
 
 
+TODO:
 
-
-
-TODO
-
-* Find better/faster way to restore locals
+  * Maybe rewrite synopsis to use fub?
